@@ -1,11 +1,12 @@
 """Function to work with dates."""
+
 import datetime
 
 
 def is_date(date: datetime.date) -> bool:
     """Check if date has only date but not time."""
     if not isinstance(date, datetime.date) or isinstance(date, datetime.datetime):
-        raise ValueError(f'should be datetime.date: {type(date)} "{date}"')
+        raise TypeError(f'should be datetime.date: {type(date)} "{date}"')
     return True
 
 
