@@ -3,10 +3,12 @@
 Note, that I only had access to an Alphabet export. I have no idea how it looks like
 for another company, or for a full profile.
 """
+
 from __future__ import annotations
 
 import csv
-from datetime import datetime
+from dataclasses import dataclass
+import datetime
 from decimal import Decimal
 from pathlib import Path
 from typing import Final
@@ -29,7 +31,7 @@ COLUMNS_RELEASE: Final[list[str]] = [
 ]
 
 COLUMNS_WITHDRAWAL: Final[list[str]] = [
-    "Date",
+    "Execution Date",
     "Order Number",
     "Plan",
     "Type",
@@ -44,9 +46,24 @@ COLUMNS_WITHDRAWAL: Final[list[str]] = [
 # These can be potentially wired through as a flag
 KNOWN_SYMBOL_DICT: Final[dict[str, str]] = {
     "GSU Class C": "GOOG",
+    "Cash": "USD",
 }
 
 
+@dataclass
+class StockSplit:
+    """Info about stock split."""
+
+    symbol: str
+    date: datetime.date
+    factor: int
+
+
+STOCK_SPLIT_INFO = [
+    StockSplit(symbol="GOOG", date=datetime.datetime(2022, 6, 15).date(), factor=20),
+]
+
+
 def _hacky_parse_decimal(decimal: str) -> Decimal:
     return Decimal(decimal.replace(",", ""))
 
@@ -59,7 +76,7 @@ def _init_from_release_report(row_raw: list[str], filename: str) -> BrokerTransa
     if row["Type"] != "Release":
         raise ParsingError(filename, f"Unknown type: {row_raw[3]}")
 
-    if row["Status"] != "Complete":
+    if row["Status"] != "Complete" and row["Status"] != "Staged":
         raise ParsingError(filename, f"Unknown status: {row_raw[5]}")
 
     if row["Price"][0] != "$":
@@ -78,7 +95,7 @@ def _init_from_release_report(row_raw: list[str], filename: str) -> BrokerTransa
     symbol = TICKER_RENAMES.get(symbol, symbol)
 
     return BrokerTransaction(
-        date=datetime.strptime(row["Vest Date"], "%d-%b-%Y").date(),
+        date=datetime.datetime.strptime(row["Vest Date"], "%d-%b-%Y").date(),
         action=ActionType.STOCK_ACTIVITY,
         symbol=symbol,
         description=row["Plan"],
@@ -91,9 +108,38 @@ def _init_from_release_report(row_raw: list[str], filename: str) -> BrokerTransa
     )
 
 
+# Morgan Stanley decided to put a notice in the end of the withdrawal report that looks
+# like that:
+# "Please note that any Alphabet share sales, transfers, or deposits that occurred on
+# or prior to the July 15, 2022 stock split are reflected in pre-split. Any sales,
+# transfers, or deposits that occurred after July 15, 2022 are in post-split values.
+# For GSU vests, your activity is displayed in post-split values."
+# It makes sense, but it totally breaks the CSV parsing
+def _is_notice(row: list[str]) -> bool:
+    return row[0][:11] == "Please note"
+
+
+def _handle_stock_split(transaction: BrokerTransaction) -> BrokerTransaction:
+    for split in STOCK_SPLIT_INFO:
+        if (
+            transaction.symbol == split.symbol
+            and transaction.action == ActionType.SELL
+            and transaction.date < split.date
+        ):
+            if transaction.quantity:
+                transaction.quantity *= split.factor
+            if transaction.price:
+                transaction.price /= split.factor
+
+    return transaction
+
+
 def _init_from_withdrawal_report(
     row_raw: list[str], filename: str
-) -> BrokerTransaction:
+) -> BrokerTransaction | None:
+    if _is_notice(row_raw):
+        return None
+
     if len(COLUMNS_WITHDRAWAL) != len(row_raw):
         raise UnexpectedColumnCountError(row_raw, len(COLUMNS_WITHDRAWAL), filename)
     row = {col: row_raw[i] for i, col in enumerate(COLUMNS_WITHDRAWAL)}
@@ -110,23 +156,32 @@ def _init_from_withdrawal_report(
     if row["Plan"] not in KNOWN_SYMBOL_DICT:
         raise ParsingError(filename, f"Unknown plan: {row_raw[3]}")
 
-    quantity = -Decimal(row["Quantity"])
+    quantity = -_hacky_parse_decimal(row["Quantity"])
     price = _hacky_parse_decimal(row["Price"][1:])
     amount = _hacky_parse_decimal(row["Net Amount"][1:])
+    fees = quantity * price - amount
 
-    return BrokerTransaction(
-        date=datetime.strptime(row["Date"], "%d-%b-%Y").date(),
-        action=ActionType.SELL,
+    if row["Plan"] == "Cash":
+        action = ActionType.TRANSFER
+        amount *= -1
+    else:
+        action = ActionType.SELL
+
+    transaction = BrokerTransaction(
+        date=datetime.datetime.strptime(row["Execution Date"], "%d-%b-%Y").date(),
+        action=action,
         symbol=KNOWN_SYMBOL_DICT[row["Plan"]],
         description=row["Plan"],
         quantity=quantity,
         price=price,
-        fees=quantity * price - amount,
+        fees=fees,
         amount=amount,
         currency="USD",
         broker="Morgan Stanley",
     )
 
+    return _handle_stock_split(transaction)
+
 
 def _validate_header(
     header: list[str], golden_header: list[str], filename: str
@@ -136,7 +191,7 @@ def _validate_header(
         raise UnexpectedColumnCountError(header, len(golden_header), filename)
     for i, (expected, actual) in enumerate(zip(golden_header, header)):
         if expected != actual:
-            msg = f"Expected column {i+1} to be {expected} but found {actual}"
+            msg = f"Expected column {i + 1} to be {expected} but found {actual}"
             raise ParsingError(filename, msg)
 
 
@@ -164,4 +219,4 @@ def read_mssb_transactions(transactions_folder: str) -> list[BrokerTransaction]:
                     _init_from_release_report(row, str(file)) for row in lines
                 ]
 
-    return transactions
+    return [transaction for transaction in transactions if transaction]
